/*
------------------------------------------------------------
Program file "mhelp.cc" Magic Help functions
Written by Paul Landes on 2/97
Modified:

Copyright (C) 1997 by Paul Landes. This code/program is free 
software; you can redistribute it and/or modify it under the terms 
of the GNU General Public License as published by the Free Software 
Foundation; either version 2 of the License, or (at your option) 
any later version. 

This header file contains function and object defs.

Unless otherwise stated at the beginning of each member function,
all functions return standard error status: 1 on fail, 0 on success.

includes:
  mhelp.h
------------------------------------------------------------
*/

#include"mhelp.h"

//this function returns the next position to place a number for the
//odd n magic squares.  current is the number to be inserted, dim is
//the dimension of the square, and the evod is even(0)/odd(1) comparitor
void getNextPos(int &row, int &col, int current, int dim)
{
  if ( (current % dim) == 0 ) row++;
  else {
    row--;
    col++;
  }
  if (row >= dim) row = 0;
  else if (row < 0) row = dim - 1;
  if (col >= dim) col = 0;
  else if (row < 0) row = dim - 1;
}

//---------(abstract)-----------MagicSquare---------------------------
//set initial data up
MagicSquare::MagicSquare()
{
  mat = 0;
  dim = 0;
  initialized = false;
  magicSum = 0;
}

//defines size of matrix, NEWs it, and sets up other importand interal
//data
int MagicSquare::start(int dimP)
{
  //if we have already initlized, don't do it again
  if ( (initialized == true) ) return 1;

  dim = dimP;
  mat = new MgPrim(dim, dim);
  magicSum = ( dim * (dim*dim + 1) ) / 2;
  initialized = true;
  return 0;
}

//check to see if all columns and rows add up to be a magic square
//(this is primarly for debugging).  Returns 2 on init error, 1 on
//fail checking rows and coloumns, and 3 on fail checking diagonals
int MagicSquare::chkValid()
{
  int acc, j, i, getit;

  if (initialized == false) return 2;

  //check rows
  for (i = 0; i < dim; i++) {
    acc = 0;
    for (j = 0; j < dim; j++) {
      mat->getElement(i, j, getit);
      acc += getit;
    }
    if (acc != magicSum) return 1;
  }

  //check columns
  for (i = 0; i < dim; i++) {
    acc = 0;
    for (j = 0; j < dim; j++) {
      mat->getElement(j, i, getit);
      acc += getit;
    }
    if (acc != magicSum) return 1;
  }

  //check main diagonals
  acc = 0;
  for (i = 0; i < dim; i++) {
    mat->getElement(i, i, getit);
    acc += getit;
  }
  if (acc != magicSum) return 3;

  acc = 0;
  for (i = 0; i < dim; i++) {
    mat->getElement(i, mat->getRow() - i - 1, getit);
    acc += getit;
  }
  if (acc != magicSum) return 3;

  return 0;
}

//swap two elements in the matrix
int MagicSquare::swap(int row1, int col1, int row2, int col2)
{
  int holder, swap;

  if (initialized == false) return 2;

  //first two statments attempt to get two elements.  If any do not
  //exist, getElement will return an error code
  if ( mat->getElement(row1, col1, swap) ) return 1;
  if ( mat->getElement(row2, col2, holder) ) return 1;

  mat->setElement(row1, col1, holder);
  mat->setElement(row2, col2, swap);
  return 0;
}


//this function is like the print function found in matrix.h.  It
//simply prints out the matrix and puts in the correct tags to make a
//standard HTML table
void MagicSquare::printHtml(bool sum)
{
  int rowcount = mat->getRow();
  int colcount = mat->getCol();
  int i, j, tmpVar;

  if (initialized == false) return;

  cout << "<table border>" << endl;
  //cell by cell print using cout
  for (i = 0; i < rowcount; i++) {
    cout << "<tr>" << endl << "        ";       //tag row
    for (j = 0; j < colcount; j++)
      {
    	mat->getElement(i, j ,tmpVar);
	cout << "<td>" << tmpVar << "</td>";   //tag data cell
      }
    cout << endl << "</tr>" << endl;           //end tag row
  }
  cout << "</table>" << endl;                  //end table

  //print out the sum if that is what the user wants
  if (sum == true) {
    cout << "<p><b>Magic Sum:</b> " << magicSum << "<br>" 
	 << endl;
  }

}

//---------------------------------MagicOdd---------------------------------
//primary algorithm to create odd matrices.  This function is used by
//other children objects, so extra parameters are introduced to make
//it more flexable. 
//offset is the number that is being inserted
//dimP is the dimention of this matrix to make
//srow/scol starting row and column
int MagicOdd::makeOddSquare(int offset, int dimP, int srow, int scol)
{
  int count = offset, row = 0, col = (dimP - 1) / 2, i;
  if ( !(dimP % 2) || ( initialized == false ) ) return 1;

  for (i = 0; i < dimP*dimP; i++) {
    mat->setElement(row + srow, col + scol, count);
    getNextPos(row, col, count, dimP);
    count++;
  }

  return 0;
}

//---------------------------------MagicSEven-------------------------------
//this is the algorithm to solve order of 4m + 2 magic squares
int MagicSEven::makeSingleEven()
{
  int quadSize, m, middleRow, corsRow, i, j;

  if ( initialized == false ) return 1;

  quadSize = dim / 2;

  //generate the 1st phase which is to do De La Loubere algorith in 4
  //evenly divided sectors of the matrix
  makeOddSquare(1, quadSize, 0, 0);
  makeOddSquare( quadSize*quadSize + 1, quadSize, 0, quadSize);
  makeOddSquare( quadSize*quadSize * 2 + 1, quadSize, quadSize, quadSize);
  makeOddSquare( quadSize*quadSize * 3 + 1, quadSize, quadSize, 0);

  //define limits
  m = (dim - 2) / 4;               //4m + 2 where m is an integer
  middleRow = (quadSize - 1) / 2;  //middle row of quad. A
  corsRow = middleRow + quadSize;  //coresponding row in quad. D

  //swap middle row of quads A and D
  for (i = 1; i <= m; i++) swap(middleRow, i, corsRow, i);
  //swap all other rows of quads A and D
  for (j = 0; j < quadSize; j++)
    if (j != middleRow)
      for (i = 0; i < m; i++) swap(j, i, j + quadSize, i);

  //swap all other rows of quads B and C
  for (j = 0; j < quadSize; j++)
    for (i = quadSize; i < quadSize + m + 2; i++) 
      swap(j, i, j + quadSize, i);

  return 0;
}

//---------------------------------MagicDEven-------------------------------
//this function swaps out the 'skew' or the 'skewed' elements to make
//a magic square
void MagicDEven::skewExchange(int m, int rowoff, int coloff)
{
  int row, col;

  for (row = rowoff; row < rowoff + m; row++) 
    for (col = coloff; col < coloff + m; col++) 
      swap(row, col, dim - row - 1, dim - col - 1);
}

//this function creats the magic square of size dimP
int MagicDEven::makeDEvenSquare(int dimP)
{
  int m, row, col, count = 0;

  //make sure everything is setup first
  if ( initialized == false ) return 1;

  //fill in consecutive numbers in matrix by row by column
  for (row = 0; row < dim; row++) 
    for (col = 0; col < dim; col++) {
      count++;
      mat->setElement(row, col, count);
    }

  //figure out our m
  m = dimP / 4;                    //4m + 2 where m is an integer

  //skew the matrix by swapping the 'skew' of the corrensponding
  //matrix
  skewExchange(m, 0, m);
  skewExchange(m, 0, 2*m);
  skewExchange(m, m, 0);
  skewExchange(m, m ,3*m);

  return 0;
}

//---------------------------------MagicDEven-------------------------------
//this is the init funciton for t
int MagicControl::init(int n)
{
  if (n <= 2) return 1;

  if (baseMat) delete baseMat;

  if ( ( (n  + 1 ) % 2) == 0) baseMat = new MagicOdd();
  else if ( ( (n - 2) % 4 ) == 0) baseMat = new MagicSEven();
  else baseMat = new MagicDEven();

  baseMat->init(n);
  return baseMat->makeMagicSquare();
}
